# -*- coding: utf-8 -*-
"""
Created on Tue Mar 10 17:22:45 2015

@author: fenet

"""

import numpy as np
import ctypes
import warnings
import os
import sys
import pylab as plt
plt.ion()


def _choose_extlib(OMP):
    """
    Returns the DLL or Shared Library that computes the (reassigned or not)
    CQgram.

    Parameters
    ----------
    OMP: if True, returns the parallel version using OpenMP.

    Returns
    -------
    ext_lib: Ctypes pointer to the DLL or SO library.
    """
    suffix = 'omp' if OMP else 'serial'
    ext = 'so' if sys.platform.startswith('linux') else 'dll'
    lib_name = 'compute_CQgram_' + suffix + '.' + ext
    return ctypes.CDLL(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                    u'bin',
                                    lib_name))


class Audio(object):
    def __init__(self, data, sr):
        self.data = data.copy()
        self.sr = float(sr)

        self.check()

    def check(self):
        if type(self.data) != np.ndarray:
            raise AttributeError('Expecting a numpy array as data')
        if self.data.dtype != np.dtype('float64'):
            raise AttributeError('Expecting a numpy array with '
                                 'double precision as data')
        if self.data.shape[1] != 1 and self.data.shape[1] != 2:
            raise AttributeError('Expecting a 2-D numpy array with '
                                 '1 column (mono signal) or '
                                 '2 columns (stereo)')
        if np.abs(self.data).max() > 1:
            raise AttributeError('Expecting a numpy array with '
                                 'values between -1 and +1')
        if type(self.sr) != float:
            warnings.warn('Expecting sampling_rate as a float. '
                          'It will be converted.')


class Spectrogram(object):
    def __init__(self, data=[], t=[], y=[], window='', hop=-1.):
        self.window = window
        self.data = data
        self.t = t
        self.y = y
        self.hop = hop
        self.check()

    def check(self):
        if type(self.window) != str:
            raise AttributeError('Window argument is expected to be a string ')
        if type(self.data) != np.ndarray:
            raise AttributeError('Expecting a numpy array as data')
        if self.data.dtype != np.dtype('float64'):
            raise AttributeError('Expecting a numpy array with '
                                 'double precision as data')
        if len(self.t) != self.data.shape[1]:
            raise AttributeError('Dimension mismatch between \'t\' and \'data\'')
        if len(self.y) != self.data.shape[0]:
            raise AttributeError('Dimension mismatch between \'y\' and \'data\'')
        if type(self.hop) != float:
            raise AttributeError('Expected \'hop\' as a float')

    def display(self):
        plt.figure()
        plt.imshow(np.log(self.data), origin='lower', aspect='auto',
                   cmap='YlOrBr', interpolation='nearest')
        plt.colorbar()


class Cqgram(Spectrogram):
    def __init__(self, data=[], r=-1., t=[], f=[], window='', hop=-1.):
        self.r = r
        super(Cqgram, self).__init__(data=data, t=t, y=f, window=window, hop=hop)

    def check(self):
        super(Cqgram, self).check()
        if type(self.r) != float:
            warnings.warn('*r* must be an integer value but should be passsed '
                                            'as float.Converting r to float')
            self.r = float(self.r)
        if int(self.r) != self.r:
            raise AttributeError('r must be an integer (number of bins per note)')


def cqgram(sound, hop, f_min, f_max, r, stereo2signal='mix', window='hann',
           OMP=True):
    """
    Returns the squared Constant-Q Spectrogram of the signal represented in
    *sound* (from the *audio* class). "Squared" means that each cell of the
    output array contains the squared magnitude of the complex scalar product
    between a local extraction of the input signal and one harmonic function.

    Parameters
    ----------
        *sound*: audio instance
            the signal of interest

        *hop*: float
            time hop (in seconds) between two successive frames
            of the CQgram

        *f_min*: float
            start frequency (in Hertz) of the CQgram

        *f_max*: float
            end frequency (in Hertz) of the CQgram

        *r*: float
            number of bins per note (typical value: 3)\n

        *stereo2signal*: 'mix', 'left' or 'right'
            in the case of a stereo sound, defines what signal is taken
            into account in the CQT calculation. *'left'* returns the CQT
            of the left channel (idem for *'right'*). *'mix'* calculates the
            CQT of the average between left and right channels. This parameter
            has no impact when processing a mono sound.

        *window*: 'rect', 'hamming' or 'hann'
            determines the type of window which is used when calculating the
            spectrum of a slice of signal

        *OMP*: set to True to use the parallel implementation using OpenMP.

    Returns
    -------
        *t_cqt*: 1D-array
            temporal indices (in seconds) of the spectrogram

        *f_cqt*: 1D-array
            frequency indices (in Hz) of the spectrogram

        *CQgram*: 2D-array of size (len(t_cqt),len(f))
            contains the spectrogram of the signal
    """

    # Pre-processing directives -----------------------------------------------
    PURE_PYTHON = 1
    # -------------------------------------------------------------------------

    # Checking input***********************************************************
    if sound.__class__.__name__ != 'Audio':
        raise AttributeError('Expecting an audio instance as 1st argument')
    if sound.data.dtype != np.dtype('float64'):
        raise AttributeError('Sound object should contain a numpy array with'
                             'double precision. Unable to proceed.')
    if type(hop) != float:
        raise AttributeError('Expecting a float as 2nd arg. (hop size in s)')
    if type(f_min) != float:
        raise AttributeError('Expecting a float as 3rd arg. (min frequency)')
    if f_min < 0 or f_min > sound.sr/2:
        raise AttributeError('f_min must be between 0 and sampling_rate/2')
    if type(f_max) != float:
        raise AttributeError('Expecting a float as 4th arguement (max freq.)')
    if f_max < f_min or f_max > sound.sr/2:
        raise AttributeError('f_max must be between Fmin and sampling_rate/2')
    if int(r) != r:
        raise AttributeError('r must be an integer (number of bins per note)')
    if window != 'hann' and window != 'hamming' and window != 'rect':
        raise AttributeError('*window* is expected to be \'hann\' or'
                             ' \'hamming\' or \'rect\' ')
    if type(OMP) != bool:
        raise AttributeError('*OMP* should be a boolean.')
    # *************************************************************************

    # Formatting input*********************************************************
    if sound.data.shape[1] == 2:
        if stereo2signal == 'left':
            data = sound.data[:, 0]
        elif stereo2signal == 'right':
            data = sound.data[:, 1]
        elif stereo2signal == 'mix':
            data = 1/2.*sound.data[:, 0] + 1/2.*sound.data[:, 1]
        else:
            raise AttributeError('stereo2signal parameter must be \'left\', '
                                                        '\'right\' or \'mix\'')
    else:
        data = sound.data[:, 0]

    bins_nbof = np.floor(np.log2(f_max/f_min)*12*r)
    # *************************************************************************

    # Computing kernel functions***********************************************
    if not hasattr(cqgram, 'kernel') or (cqgram.input_sr != sound.sr or
                                         cqgram.f_min != f_min or
                                         cqgram.f_max != f_max or
                                         cqgram.r != r or
                                         cqgram.window != window):
        print('1st call => Calculation of the Spectral Kernels required')
        # Setting of the static variables
        cqgram.input_sr = sound.sr
        cqgram.f_min = f_min
        cqgram.f_max = f_max
        cqgram.r = r
        cqgram.window = window
        # Calculation of the Harmonic Kernels
        cqgram.f_cqt = np.zeros(bins_nbof)
        for k in range(int(bins_nbof)):
            cqgram.f_cqt[k] = f_min*2**(float(k)/(12*r))
            delta_f_k = (2.**(1./(12.*r))-1.)*cqgram.f_cqt[k]
            N_k = np.floor((1./delta_f_k)*sound.sr)
            if window == 'rect':
                window_k = np.ones(N_k)
            elif window == 'hann':
                window_k = np.hanning(N_k)
            elif window == 'hamming':
                window_k = np.hamming(N_k)
            else:
                raise SystemError('PROBLEM IN THE CODE, unexpected & forbidden'
                                                            'value for window')
            if k == 0:
                cqgram.frame_len = N_k
                cqgram.kernel = np.zeros([bins_nbof, N_k], dtype=np.complex128)
                cqgram.kersupp = np.zeros([bins_nbof, 2], dtype=np.float64)

            t_red = np.linspace(0, N_k-1, N_k)
            harmonic_func = np.exp(-2j*np.pi*cqgram.f_cqt[k]/sound.sr*t_red)
            cqgram.kernel[k] = np.pad(window_k*harmonic_func,
                                   (np.floor((cqgram.frame_len - N_k)/2),
                                    cqgram.frame_len -
                                    (N_k+np.floor((cqgram.frame_len - N_k)/2))),
                                   'constant')/N_k
            cqgram.kersupp[k] = [np.floor((cqgram.frame_len - N_k)/2),
                                 np.floor((cqgram.frame_len - N_k)/2)+N_k-1]
    # *************************************************************************

    # Allocating memory *******************************************************
    CQgram_len = np.ceil((len(data)-cqgram.frame_len+1)/np.floor(hop*sound.sr))
    CQgram = np.zeros([bins_nbof, CQgram_len])
    t_cqt = (cqgram.frame_len/2.+np.arange(0, len(data)-cqgram.frame_len+1, np.floor(hop*sound.sr)))/sound.sr
    # *************************************************************************

    # Calculating CQgram ******************************************************
    if PURE_PYTHON == True:
        j = 0.
        j_red = 0
        while j+cqgram.frame_len-1 < len(data):
            signal = data[j:j+cqgram.frame_len]
            for i in range(int(bins_nbof)):
                CQgram[i, j_red] = abs(sum(signal[cqgram.kersupp[i,0]:cqgram.kersupp[i,1]+1]*cqgram.kernel[i,cqgram.kersupp[i,0]:cqgram.kersupp[i,1]+1]))**2
            j += np.floor(hop*sound.sr)
            j_red += 1
    else:
        ext_lib = _choose_extlib(OMP)
        ext_lib.compute_CQgram(data.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               cqgram.kernel.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               cqgram.kersupp.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               CQgram.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), #output of C++ code
                               ctypes.c_int(int(cqgram.frame_len)),
                               ctypes.c_int(int(CQgram_len)),
                               ctypes.c_int(int(bins_nbof)),
                               ctypes.c_int(int(len(data))),
                               ctypes.c_int(int(np.floor(hop*sound.sr)))
                               )
    # *************************************************************************

    return Cqgram(data=CQgram, r=r, t=t_cqt, f=cqgram.f_cqt)


# *****************************************************************************
# *****************************************************************************
# ****************************Reassigned CQgram********************************
# *****************************************************************************
# *****************************************************************************


def rcqgram(sound, hop, f_min, f_max, r, stereo2signal='mix', window='hann',
            filt_ord=500, OMP=True):
    """
    Returns the squared Reassigned Constant-Q Spectrogram of the signal
    represented in *sound* (from the *audio* class). "Squared" means that
    the content of one cell of the output array is the sum of the squared
    magnitudes of the harmonic projections assigned to this cell.

    Parameters
    ----------
        *sound*: audio instance
            the signal of interest

        *hop*: float
            time hop (in seconds) between two successive frames
            of the CQgram

        *f_min*: float
            start frequency (in Hertz) of the CQgram

        *f_max*: float
            end frequency (in Hertz) of the CQgram

        *r*: float
            number of bins per note (typical value: 3)\n

        *stereo2signal*: 'mix', 'left' or 'right'
            in the case of a stereo sound, defines what signal is taken
            into account in the CQT calculation. *'left'* returns the CQT
            of the left channel (idem for *'right'*). *'mix'* calculates the
            CQT of the average between left and right channels. This parameter
            has no impact when processing a mono sound.

        *window*: 'rect', 'hamming' or 'hann'
            determines the type of window which is used when calculating the
            spectrum of a slice of signal

        *filt_ord*: gives the order of the ideal differentiator filter.
            Beware that the length of the filter will
            actually be 2*filt_ord+1.

        *OMP*: set to True to use the parallel implementation using OpenMP.


    Returns
    -------
        *t_cqt*: 1D-array
            temporal indices (in seconds) of the spectrogram

        *f_cqt*: 1D-array
            frequency indices (in Hz) of the spectrogram

        *CQgram*: 2D-array of size (len(t_cqt),len(f))
            contains the spectrogram of the signal
    """

    # Setting "pre-processing" directives**************************************
    PURE_PYTHON = 1
    # *************************************************************************

    # Checking input***********************************************************
    if sound.__class__.__name__ != 'Audio':
        raise AttributeError('Expecting an audio instance as 1st argument')
    if sound.data.dtype != np.dtype('float64'):
        raise AttributeError('Sound object should contain a numpy array with'
                             'double precision. Unable to proceed.')
    if type(hop) != float:
        raise AttributeError('Expecting a float as 2nd arg. (hop size in s)')
    if type(f_min) != float:
        raise AttributeError('Expecting a float as 3rd arg. (min frequency)')
    if f_min < 0 or f_min > sound.sr/2:
        raise AttributeError('f_min must be between 0 and sampling_rate/2')
    if type(f_max) != float:
        raise AttributeError('Expecting a float as 4th arguement (max freq.)')
    if f_max < f_min or f_max > sound.sr/2:
        raise AttributeError('f_max must be between Fmin and sampling_rate/2')
    if int(r) != r:
        raise AttributeError('r must be an integer (number of bins per note)')
    if window != 'hann' and window != 'hamming' and window != 'rect':
        raise AttributeError('window is expected to be \'hann\' or'
                             ' \'hamming\' or \'rect\' ')
    if type(filt_ord) != int:
        raise AttributeError('Expecting an int as filt_ord')
    if type(OMP) != bool:
        raise AttributeError('*OMP* should be a boolean.')
    # *************************************************************************

    # Formatting input*********************************************************
    if sound.data.shape[1] == 2:
        if stereo2signal == 'left':
            data = sound.data[:, 0]
        elif stereo2signal == 'right':
            data = sound.data[:, 1]
        elif stereo2signal == 'mix':
            data = 1/2.*sound.data[:, 0] + 1/2.*sound.data[:, 1]
        else:
            raise AttributeError('stereo2signal parameter must be \'left\', '
                                                        '\'right\' or \'mix\'')
    else:
        data = sound.data[:, 0]

    bins_nbof = np.floor(np.log2(f_max/f_min)*12*r)
    # *************************************************************************

    # Computing kernel functions **********************************************
    if not hasattr(rcqgram, 'kernel') or (rcqgram.input_sr != sound.sr or
                                          rcqgram.f_min != f_min or
                                          rcqgram.f_max != f_max or
                                          rcqgram.r != r or
                                          rcqgram.window != window):
        print('1st call => Calculation of the Spectral Kernels required')
        # Setting of the static variables
        rcqgram.input_sr = sound.sr
        rcqgram.f_min = f_min
        rcqgram.f_max = f_max
        rcqgram.r = r
        rcqgram.window = window
        # Calculation of the Harmonic Kernels
        rcqgram.f_cqt = np.zeros(bins_nbof)
        for k in range(int(bins_nbof)):
            rcqgram.f_cqt[k] = f_min*2**(float(k)/(12*r))
            delta_f_k = (2.**(1./(12.*r))-1.)*rcqgram.f_cqt[k]
            N_k = np.floor((1./delta_f_k)*sound.sr)
            if window == 'rect':
                window_k = np.ones(N_k)
            elif window == 'hann':
                window_k = np.hanning(N_k)
            elif window == 'hamming':
                window_k = np.hamming(N_k)
            else:
                raise SystemError('PROBLEM IN THE CODE, unexpected & forbidden'
                                                            'value for window')

            u = np.arange(-filt_ord, filt_ord+1, 1, dtype=np.double)
            g = (-1)**u/u
            g[filt_ord] = 0
            window_k_diff = np.convolve(window_k, g)

            N_prim_k = len(window_k_diff)
            supp_size_k = N_prim_k  # /!\
            window_k = np.pad(window_k, (np.floor((supp_size_k - N_k)/2),
                                        supp_size_k - N_k - np.floor((supp_size_k-N_k)/2)),'constant')
            window_k_diff = np.pad(window_k_diff, (np.floor((supp_size_k - N_prim_k)/2),
                                        supp_size_k - N_prim_k - np.floor((supp_size_k-N_prim_k)/2)),'constant')

            if k == 0:
                rcqgram.frame_len = supp_size_k
                rcqgram.kernel = np.zeros([bins_nbof, rcqgram.frame_len], dtype=np.complex128)
                rcqgram.dkernel = np.zeros([bins_nbof, rcqgram.frame_len], dtype=np.complex128)
                rcqgram.ikernel = np.zeros([bins_nbof, rcqgram.frame_len], dtype=np.complex128)                
                rcqgram.kersupp = np.zeros([bins_nbof, 2], dtype=np.float64)
 
            t_centered = np.linspace(-np.floor(supp_size_k/2),
                                         -np.floor(supp_size_k/2)+supp_size_k-1, supp_size_k)
            harmonic_func = np.exp(-2j*np.pi*rcqgram.f_cqt[k]/sound.sr*t_centered)
            rcqgram.kernel[k] = np.pad(window_k*harmonic_func,
                                   (np.floor((rcqgram.frame_len - supp_size_k)/2),
                                    rcqgram.frame_len -
                                    (supp_size_k+np.floor((rcqgram.frame_len - supp_size_k)/2))),
                                   'constant')/N_k
            rcqgram.dkernel[k] = np.pad(window_k_diff*harmonic_func,
                                   (np.floor((rcqgram.frame_len - supp_size_k)/2),
                                    rcqgram.frame_len -
                                    (supp_size_k+np.floor((rcqgram.frame_len - supp_size_k)/2))),
                                    'constant')/N_k
            rcqgram.ikernel[k] = np.pad(t_centered/sound.sr*window_k*harmonic_func,
                                   (np.floor((rcqgram.frame_len - supp_size_k)/2),
                                    rcqgram.frame_len -
                                    (supp_size_k+np.floor((rcqgram.frame_len - supp_size_k)/2))),
                                   'constant')/N_k
            rcqgram.kersupp[k] = [np.floor((rcqgram.frame_len - supp_size_k)/2),
                                  np.floor((rcqgram.frame_len - supp_size_k)/2)+N_prim_k-1]
    # *************************************************************************

    # Allocating memory *******************************************************
    CQgram_len = np.ceil((len(data)-rcqgram.frame_len+1)/np.floor(hop*sound.sr))
    CQgram = np.zeros([bins_nbof, CQgram_len])
    t_cqt = (rcqgram.frame_len/2.+np.arange(0, len(data)-rcqgram.frame_len+1, np.floor(hop*sound.sr)))/sound.sr
    # *************************************************************************

    # Calculating Reassigned CQgram *******************************************
    if PURE_PYTHON == True:
        j = 0.
        j_red = 0
        while j+rcqgram.frame_len-1 < len(data):
            signal = data[j:j+rcqgram.frame_len]
            for i in range(int(bins_nbof)):
                CQcomplex_i_jred = sum(signal[rcqgram.kersupp[i,0]:rcqgram.kersupp[i,1]+1]*rcqgram.kernel[i,rcqgram.kersupp[i,0]:rcqgram.kersupp[i,1]+1])

                delta_t = np.real(sum(rcqgram.ikernel[i,rcqgram.kersupp[i,0]:rcqgram.kersupp[i,1]+1]         
                                * signal[rcqgram.kersupp[i,0]:rcqgram.kersupp[i,1]+1]) /
                                                                            CQcomplex_i_jred )                
                delta_f = sound.sr/(2*np.pi)*np.imag(sum(rcqgram.dkernel[i,rcqgram.kersupp[i,0]:rcqgram.kersupp[i,1]+1]
                                *signal[rcqgram.kersupp[i,0]:rcqgram.kersupp[i,1]+1]) / 
                                                                            CQcomplex_i_jred )            

                if(abs(delta_t) > rcqgram.frame_len/(2*sound.sr)):
                #When the CQgram value is too low, it may generate reassignment
                #artefacts (division by low value). We prevent these by 
                #checking that the reassigned time index lies within the bounds 
                #of the frame of signal.
                    delta_t = 0
                    delta_f = 0

                t_rea = t_cqt[j_red]+delta_t
                j_rea = np.round((t_rea-t_cqt[0]) / (t_cqt[1]-t_cqt[0]))
                f_rea = rcqgram.f_cqt[i]-delta_f
                i_rea = np.round(np.log(f_rea/rcqgram.f_cqt[0]) /
                                    np.log(rcqgram.f_cqt[1]/rcqgram.f_cqt[0]))
                if i_rea >= 0 and i_rea < CQgram.shape[0] and j_rea >= 0 and j_rea < CQgram.shape[1]:
                    CQgram[i_rea, j_rea] += CQcomplex_i_jred*CQcomplex_i_jred.conjugate()

            j += np.floor(hop*sound.sr)
            j_red += 1
    else:
        ext_lib = _choose_extlib(OMP)
        ikernel_samples = rcqgram.ikernel*sound.sr
        nu_cqt = rcqgram.f_cqt/sound.sr
        ext_lib.compute_rCQgram(data.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               rcqgram.kernel.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               ikernel_samples.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), #the C++ code is made independant of the sampling freq                              
                               rcqgram.dkernel.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               rcqgram.kersupp.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               nu_cqt.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),  #the C++ code is made independant of the sampling freq
                               CQgram.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), #output of C++ code
                               ctypes.c_int(int(rcqgram.frame_len)),
                               ctypes.c_int(int(CQgram_len)),
                               ctypes.c_int(int(bins_nbof)),
                               ctypes.c_int(int(len(data))),
                               ctypes.c_int(int(np.floor(hop*sound.sr)))
                               )      
    # *************************************************************************

    return Cqgram(data=CQgram, r=r, t=t_cqt, f=rcqgram.f_cqt, window=window,
                  hop=hop)
